# Non-linear Ordinal Logistic Regression
#
# Mark Mathieson   mathies@stats.ox.ac.uk
# University of Oxford, UK.
# Based on the network code of B.D.Ripley
#
# Version October 1996

nolr <- function(x, y, size, weights, Wts, skip=T,
		 riskfit=F, loss=0, phi,
		 decay=0.0, cutreg=F, thetamean=5.0, thetavar=2.0, rang=1.0,
		 trace=T, maxit=7000, maxitb=0, mask, 
		 mu=0.0, gini=0.0, dgini=0.0, dg, ...)
{
  if(!is.loaded(symbol.C("set_nolr")))
    dyn.load("nolr.o")  

  net <- NULL
  linout <- T
  k <- max(y)
  n <- nrow(x)
  x <- as.matrix(x)
  y <- as.matrix(y)
  net$k <- k
  net$cutreg <- cutreg
  net$riskfit <- riskfit
  net$skip <- skip

  if(length(which.na(x)))
    stop("missing values in x")
  if(length(which.na(y)))
    stop("missing values in y")
  if(dim(x)[1] != dim(y)[1])
    stop("nrows of x and y must match")
  if((!missing(phi)) && (length(phi)!=(k-1)))
    stop("number of cutpoints doesn't match number of classes")
  if(riskfit && missing(loss))
    stop("must specify loss matrix for risk fit")
  if(riskfit && ((nrow(loss)!=k) || (ncol(loss)!=k)))
    stop(cat("loss matrix must be square matrix of size",k))
  if(nrow(x)!=nrow(y))
    stop("x and y are of different lengths")
#  if(!all(sort(unique(y))==(1:k)))
#    stop(cat("not all of categories 1 to ",max(y)," are represented"))
  if((thetamean<0.0) || (thetavar<0.0))
    stop(cat("thetamean and thetavar must both be positive"))
  if(riskfit && (ncol(y)>1))
    stop("censoring only works with fitting by maximum likelihood")
  if((ncol(y)>2))
    stop("class vector is incorrect")

  # Check censoring

  if (ncol(y)==1)
    y <- cbind((y-1),y)
    else
      y[,1] <- y[,1] - 1
  
  y <- cbind(y[,2],y[,1])
  if (any((y[,1]-y[,2])<1))
    stop("censoring incorrect")

  # Cut-point regularization

  if(!cutreg) {
    lambda <- 0.0
    alpha <- 1.0
  } else {
    alpha <- (thetamean*thetamean)/thetavar
    lambda <- thetamean / thetavar
    net$lambda <- lambda
    net$alpha <- alpha
  }

  # Risk fitting

  if(!riskfit) loss <- 0.0
  net$loss <- loss

  # Setup cutpoints

  if(missing(phi)) {
    phi <- rep(0.0,(k-1))
    phi[1] <- runif(1,-1.0,0.0)
    for (i in 2:(k-1))
      phi[i] <- phi[i-1] + runif(1,2.0,4.0)
  }
  if(any(diff(phi))<=0.0)
    stop("cut-points must be monotonic increasing")
  net$phi <- phi

  # Setup connections

  net$n <- c(dim(x)[2], size, 1)
  net$nunits <- 1 + sum(net$n)
  net$nconn <- rep(0, net$nunits+1)
  net$conn <- numeric(0)
  net <- norm.nolr(net)
  if(skip) net <- add.nolr(net, seq(1,net$n[1]), 
			  seq(1+net$n[1]+net$n[2], net$nunits-1))
  nsunits <- net$nunits
  if(linout) nsunits <- net$nunits - net$n[3]
  net$nsunits <- nsunits

  # Setup connection weights

  if(length(net$conn)==0) stop("no weights to fit")
  if(missing(Wts))
    if(rang > 0)
      wts <- runif(length(net$conn), -rang, rang)
      else wts <- rep(0, length(net$conn))
    else wts <- Wts
  if(length(wts) != length(net$conn))
    stop("connection weights vector of incorrect length")
  net$wts <- wts

  # Setup decay

  if(length(decay)==1) decay <- rep(decay,length(net$wts))
  if(length(decay)!=length(net$wts))
    stop("incorrect length of decay vector")
  net$decay <- decay
  net$mu <- mu
  net$gini <- gini
  net$dgini <- dgini

  # Initialize

  .C("set_nolr",
     as.integer(net$n),
     as.integer(net$nconn),
     as.integer(net$conn),
     as.single(net$decay),
     as.integer(net$nsunits),
     as.integer(net$skip),
     as.integer(net$k),
     as.integer(net$riskfit),
     as.single(as.vector(net$loss)),
     as.integer(net$cutreg),
     as.single(net$lambda),
     as.single(net$alpha),
     as.single(net$mu),
     as.single(net$gini),
     as.single(net$dgini)
    )

  # Setup sample weights

  ntr <- dim(x)[1]
  if(missing(weights))
    weights <- rep(1, ntr)
  if(length(weights) != ntr || any(weights < 0))
    stop("invalid sample weights vector")
  net$weights <- weights

  # Setup DGini weights

  if(missing(dg))
    dg <- rep(0, ntr)
  if(length(dg) != ntr)
    stop("invalid sample DGini `dg' vector")
  net$dg <- dg

  z <- .C("set_otrain",
	  as.integer(ntr),
	  as.single(cbind(x,y)),
	  as.single(net$weights),
	  as.single(net$dg)
	  )

  # Optimize fixing cutpoints

  params <- c(net$wts,net$phi)
  if (maxitb!=0) {
    wt.mask <- c(rep(1,length(net$wts)+1),rep(0,(k-2)))
    m <- .C("dovm_nolr",
	    Np = as.integer(length(net$wts)+k-1),
	    params = as.single(params),
	    val = double(1),
	    maxit = as.integer(maxitb),
	    trace = trace,
	    mask = as.integer(wt.mask)
	    )
    params <- m$params
  }

  # Optimize all parameters

  if(missing(mask))
    mask <- rep(1,length(params))
  m <- .C("dovm_nolr",
	  Np = as.integer(length(net$wts)+k-1),
	  params = as.single(params),
	  val = double(1),
	  maxit = as.integer(maxit),
	  trace = trace,
	  mask = as.integer(mask)
	  )

  net$E <- m$val
  net$wts <- m$params[1:length(net$wts)]
  net$phi <- m$params[(length(net$wts)+1):length(m$params)]

  return(structure(net, class="nolr", numerics=m))
}

set.otrain <- function(x,y,weights,dg)
{
  x <- as.matrix(x)
  y <- as.matrix(y)

  if (ncol(y)==1)
    y <- cbind((y-1),y)  
  y <- cbind(y[,2],y[,1])
  if (any((y[,1]-y[,2])<1))
    stop("censoring incorrect")

  if (missing(weights))
    weights <- rep(1.0,nrow(x))
  if (missing(dg))
    dg <- rep(0.0,nrow(x))

  .C("set_otrain",
     as.integer(nrow(x)),
     as.single(cbind(x,y)),
     as.single(weights),
     as.single(dg))
  invisible(NULL)
}

set.oparams <- function(wts,phi)
{
  .C("set_parameters",
     as.single(wts),
     as.single(phi))
  invisible(NULL)
}

set.nolr <- function(net)
{
  .C("set_nolr",
     as.integer(net$n),
     as.integer(net$nconn),
     as.integer(net$conn),
     as.single(net$decay),
     as.integer(net$nsunits),
     as.integer(net$skip),
     as.integer(net$k),
     as.integer(net$riskfit),
     as.single(as.vector(net$loss)),
     as.integer(net$cutreg),
     as.single(net$lambda),
     as.single(net$alpha),
     as.single(net$mu),
     as.single(net$gini),
     as.single(net$dgini))
  invisible(NULL)
}

predict.nolr <- function(net, x, wts, phi, type="predictor")
{
  if(!is.loaded(symbol.C("set_nolr")))
    dyn.load("nolr.o")

  if(!inherits(net, "nolr"))
    stop("object not of class nolr")

  if(!any(type==c("probs","predictor")))
    stop("not a valid value for type")
 
  set.nolr(net)

  if(missing(wts))
    wts <- net$wts
  if(missing(phi))
    phi <- net$phi

  set.oparams(wts,phi)
  
  if(is.null(dim(x)))
    dim(x) <- c(1, length(x))
  if(length(which.na(x)))
    stop("missing values in x")
  x <- as.matrix(x)
  ntr <- dim(x)[1]
  nout <- net$n[3]
  
  z <- .C("onntest",
	  as.integer(ntr),
	  as.single(x),
	  tclass = single(ntr*nout)
	  )
  z <- matrix(z$tclass, ntr, nout)

  dimnames(z) <- list(dimnames(x)[[1]], dimnames(net$fitted)[[2]])
  z <- as.vector(z)
  
  if (type=="predictor")
    return(z)
  else
    return(t(sapply(z, probs.olr, phi=c(-Inf, net$phi, Inf) )))
}

# Posterior olr probabilities
probs.olr <- function(eta,phi)
{
  return(sigmoid(phi[-1]-eta) - sigmoid(phi[-length(phi)]-eta))
}


nolr.func <- function(p)
{
  z <- .C("Dfminfn",
	  val = double(1),
	  p = as.single(p)
	  )
  return(z$val)
}


nolr.fn <- function(wts,phi)
{
  p <- c(wts,phi)
  z <- .C("Dfminfn",
	  val = double(1),
	  p = as.single(p)
	  )
  return(z$val)
}

nolr.gr <- function(wts,phi)
{
  p <- c(wts,phi)
  z <- .C("Dfmingr",
	  p = as.single(p),
	  df = as.single(rep(0.0,length(p)))
	  )
  return(z$df)
}
  

add.nolr <- function(net,from, to)
{
  nconn <- net$nconn
  conn <- net$conn
  for(i in to){
    ns <- nconn[i+2]
    cadd <- from
    if((nconn[i+1] == ns) && (i<(net$n[1]+net$n[2]+1))) cadd<-c(0,from)
    con <- NULL
    if(ns > 1) con <- conn[1:ns]
    con <- c(con, cadd)
    if(length(conn) > ns) 
      con <- c(con, conn[(ns+1):length(conn)])
    for(j in (i+1):net$nunits) 
      nconn[j+1] <- nconn[j+1]+length(cadd)
    conn <- con
  }
  net$nconn <- nconn
  net$conn <- con
  net
}

norm.nolr <- function(net)
{
  n<- net$n; n0 <- n[1]; n1 <- n0+n[2]; n2 <- n1+n[3];
  if(n[2] <= 0) return(net)
  net <- add.nolr(net, 1:n0,(n0+1):n1)
  add.nolr(net, (n0+1):n1, (n1+1):n2)
}

Hess.nolr <- function(net, x, y, weights)
{
  if((!missing(x) && missing(y)) ||
     (missing(x) && !missing(y)))
    stop("Supply both or neither of x and y")

  if(!is.loaded(symbol.C("set_nolr")))
    dyn.load("nolr.o")
  nw <- length(net$wts)
  k <- net$k

  set.nolr(net)

  if (!missing(x) && !missing(y)) {
    if(missing(weights))
      weights <- rep(1, nrow(x))
    set.otrain(x,y,weights)
  }

  m <- .C("nolrHessian",
	  inwts = as.single(net$wts),
	  phi = as.single(net$phi),
	  Hess = single((nw+k-1)*(nw+k-1))
	  )
  return(matrix(m$Hess, nrow=(nw+k-1), ncol=(nw+k-1)))
}

print.nolr <- function(net,...)
{
  if(!inherits(net, "nolr"))
    stop("Not a legitimate neural net fit for nolr")
  cat("a ",net$n[1],"-",net$n[2],"-",net$n[3]," network for nolr", sep="")
  cat(" with", length(net$wts),"weights for",net$k,"classes\n")
  cat("options were -")
  tconn <- diff(net$nconn)
  if(tconn[length(tconn)] > net$n[2]+1) cat(" skip-layer connections\n")
  if(net$nunits > net$nsunits) cat("linear output units\n")
  if(any(net$decay!=0)) {
    cat("Weight decay ");
    if(all(net$decay==net$decay[1])) {
    cat("=", net$decay[1],"\n")
  } else {
    cat("- non-uniform\n")
  }}
  if(net$cutreg) cat("Cut-point regularization: mean=",net$alpha/net$lambda,
		     " var=",net$alpha/(net$lambda*net$lambda),"\n",sep="")
  if(net$riskfit) cat("Risk fit\n")
  if(!net$riskfit) cat("Parameters estimated by ML\n")
  cat("-log(L) = ",net$E,"\n")
  cat("\n")
}

summary.nolr <- function(net,...)
{
  if(!inherits(net, "nolr"))
    stop("Not a legitimate neural net fit for nolr")
  cat("a ",net$n[1],"-",net$n[2],"-",net$n[3]," network for nolr", sep="")
  cat(" with", length(net$wts),"weights for",net$k,"classes\n")
  cat("options were -")
  tconn <- diff(net$nconn)
  if(tconn[length(tconn)] > net$n[2]+1) cat(" skip-layer connections\n")
  if(net$nunits > net$nsunits) cat("linear output units\n")
  non.unif <- F
  if(any(net$decay!=0)) {
    cat("Weight decay ");
    if(all(net$decay==net$decay[1])) {
    cat("=", net$decay[1],"\n")
  } else {
    non.unif <- T
    cat("- non-uniform\n")
  }}
  if(net$cutreg) cat("Cut-point regularization: mean=",net$alpha/net$lambda,
		     " var=",net$alpha/(net$lambda*net$lambda),"\n",sep="")
  if(net$riskfit) {
    cat("Risk fit with loss matrix:\n")
    print(net$loss)
  }
  if(!net$riskfit) cat("Parameters estimated by maximum likelihood")
  cat("\nWeights (decays):\n")
  wts <- format(round(net$wts,2))
  names(wts) <- apply(cbind(net$conn, rep(1:net$nunits - 1, tconn)), 
		      1, function(x)  paste(x, collapse = "->"))
  if(non.unif)
    names(wts) <- paste(names(wts)," (",format(round(net$decay,2)),")",sep="")
   
  print(wts, quote = F)
  cat("\nCut-points:\n")
  print(net$phi)
  cat("\n")
  cat("-log(L) = ",net$E,"\n")
  invisible()
}

# Network functions
sigmoid <- function(x)
{
  return(1/(1+exp(-1*x)))
}

logit <- function(x)
{
  return(log(x/(1-x)))
}
